/**
 * JORM: an implementation of a generic mapping system for persistent Java
 *       objects. Two mapping are supported: to RDBMS and to binary files.
 * Copyright (C) 2001-2003 France Telecom R&D - INRIA
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Contact: jorm-team@objectweb.org
 *
 */

package org.objectweb.jorm.mapper.rdb.adapter;

import org.objectweb.jorm.mapper.rdb.adapter.api.RdbAdapterException;
import java.sql.ResultSet;
import java.sql.SQLException;
import java.sql.Connection;

/**
 * @author	Tom Lane
 */

public class CloudscapeAdapter extends BasicRdbAdapter {
	private final static String CLOUDSCAPESUBSTRING = "substr";
	private final static String CLOUDSCAPECONCAT = "||";
	private final static String CLOUDSCAPEFIRSTLOCATE = "locate";

	public CloudscapeAdapter() {
		super("cloudscape");
	}

	public boolean supportBatchPreparedStatement() {
		return false;
	}

	public String getSqlType(int typeCode, boolean usedInPK, int size, int scale) throws RdbAdapterException {
		switch (typeCode) {
		case TYPECODE_BOOLEAN:
		case TYPECODE_OBJBOOLEAN:
			return "SMALLINT";
		case TYPECODE_BYTE:
		case TYPECODE_OBJBYTE:
			return "SMALLINT";
		case TYPECODE_STRING:
		case TYPECODE_CHARARRAY:
		    if (size != NOSIZE) {
			  return "VARCHAR(" + size + ")";
		    }
		    // LONG VARCHAR might look like the thing to use here,
		    // but it's a useless datatype because no comparison ability
		    return "VARCHAR(256)";
		case TYPECODE_BYTEARRAY:
		case TYPECODE_SERIALIZED:
		    if (size != NOSIZE) {
			  return "VARCHAR(" + size + ") FOR BIT DATA";
		    }
		    return "VARCHAR(256) FOR BIT DATA";
		case TYPECODE_BIGINTEGER:
			if (size != NOSIZE) {
				return "DECIMAL(" + size + ",0)";
			}
			return "DECIMAL(31,0)";
		case TYPECODE_BIGDECIMAL:
			if (size != NOSIZE) {
				if (scale != NOSIZE) {
					return "DECIMAL(" + size + ", " + scale + ")";
				} else {
					return "DECIMAL(" + size + ", 4)";
				}
			} else {
				return "DECIMAL(31,4)";
			}
		default:
			return super.getSqlType(typeCode, usedInPK, size, scale);
		}
	}

	public String getValueAsSQLString(Object value, int typeCode) {
		if (value == null)
			return null;
		switch (typeCode) {
		case TYPECODE_BIGINTEGER:
		case TYPECODE_BIGDECIMAL:
			return value.toString();
		default:
			return super.getValueAsSQLString(value, typeCode);
		}
	}

	public String getSubstringName() {
		return CLOUDSCAPESUBSTRING;
	}

	public String getConcatExpression(String op1, String op2) {
		return op1 + CLOUDSCAPECONCAT + op2;
	}

	public String getFirstLocateExpression(String substring, String instring) {
		return CLOUDSCAPEFIRSTLOCATE + "(" + substring + ", " + instring + ")";
	}

	public String getIndexedLocateExpression(String instring, String substring,
											 String fromIndex) {
		return CLOUDSCAPEFIRSTLOCATE + "(" + substring + ", " + instring
				+ ", " + fromIndex + ")";
	}

    // overrides existRelation for getTables with table names not supported by driver
    protected boolean existRelation(Connection connection, String relName, String[] relationTypes) throws SQLException {
        boolean existtable = false;
        ResultSet rs = null;
        if (relationTypes == null) {
            relationTypes = new String[]{"TABLE", "VIEW"};
        }
        
        try {
            rs = connection.getMetaData().getTables(null, null, null, relationTypes);
            while (rs.next() && !existtable) {
                existtable = relName.equalsIgnoreCase(rs.getString(3));
            }
        } finally {
            if (rs != null) {
                rs.close();
            }
        }
        return existtable;
    }

	public String getSubqueryAlias() {
		return " as foo";
	}

    public String getCreateSequence(String seqName)
	throws UnsupportedOperationException {
	throw new UnsupportedOperationException("Cloudscape does not support SEQUENCE") ;
    }
    public String getCreateSequence(String seqName, Integer startid, Integer inc, Integer cache) {
	throw new UnsupportedOperationException("Cloudscape does not support SEQUENCE") ;
    }

}
