/**
 * JORM: an implementation of a generic mapping system for persistent Java
 *       objects. Two mapping are supported: to RDBMS and to binary files.
 * Copyright (C) 2001-2003 France Telecom R&D - INRIA
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Contact: jorm-team@objectweb.org
 *
 */

package org.objectweb.jorm.mapper.rdb.adapter;

import org.objectweb.jorm.mapper.rdb.adapter.api.RdbAdapter;
import org.objectweb.jorm.mapper.rdb.adapter.api.RdbAdapterException;

import javax.sql.DataSource;
import java.sql.Connection;
import java.sql.SQLException;
import java.util.Collections;
import java.util.Map;
import java.util.Properties;
import java.util.StringTokenizer;

/**
 * This factory permits to create new instances of RdbAdapter.
 * @author	S. Chassande-Barrioz
 */
public class RdbAdapterFactory {
	/**
	 * This constant represents the data base name for a standard jdbc database
	 */
	public static final String DATABASE_NAME_JDBC = "jdbc";

	/**
	 * Map between DatabaseProductName to adapter class name
	 */
	private static Properties product2adapter;

	static {
		product2adapter = new Properties();

		//Register the rdb adapter class name for each well known database
		// product name (response to Connection.getDatabaseProductName())
		product2adapter.setProperty("PostgreSQL",
			"org.objectweb.jorm.mapper.rdb.adapter.PostgresAdapter");
		product2adapter.setProperty("DB2",
			"org.objectweb.jorm.mapper.rdb.adapter.Db2Adapter");
		product2adapter.setProperty("FireBird",
			"org.objectweb.jorm.mapper.rdb.adapter.FireBirdAdapter");
		product2adapter.setProperty("Hsql",
			"org.objectweb.jorm.mapper.rdb.adapter.HsqlAdapter");
		product2adapter.setProperty("McKOI",
			"org.objectweb.jorm.mapper.rdb.adapter.MckoiAdapter");
		product2adapter.setProperty("MySQL",
			"org.objectweb.jorm.mapper.rdb.adapter.MysqlAdapter");
		product2adapter.setProperty("Oracle",
			"org.objectweb.jorm.mapper.rdb.adapter.OracleAdapter");
		product2adapter.setProperty("SAPDB",
			"org.objectweb.jorm.mapper.rdb.adapter.SapdbAdapter");
		product2adapter.setProperty("Microsft SQL Server",
			"org.objectweb.jorm.mapper.rdb.adapter.SqlAdapter");
		product2adapter.setProperty("Sybase",
			"org.objectweb.jorm.mapper.rdb.adapter.SybaseAdapter");
        product2adapter.setProperty("PROGRESS",
        	"org.objectweb.jorm.mapper.rdb.adapter.ProgressAdapter");

		//parse a system property in order add use specialization
		String str = System.getProperty("org.objectweb.jorm.mapper.rdb.adapters");
		if (str != null && str.length() >0) {
			StringTokenizer st = new StringTokenizer(str, ",=:", true);
			String first = null;
			while(st.hasMoreElements()) {
				String token = st.nextToken();
				if (token.equals("=") || token.equals(":")) {
					if (first == null) {
						System.out.println("Malformed option: " + str);
						break;
					}
				} else if (token.equals(",")) {
					first = null;
				} else if (first == null) {
					first = token;
				} else {
					product2adapter.setProperty(first, token);
					first = null;
				}
			}
		}
	}

	public static Map getProductName2Adapter() {
		return Collections.unmodifiableMap(product2adapter);
	}

	/**
	 * This first possibility permits to specify the name for which a
	 * RdbAdapter is needed.
	 * @param dbn : the data base name
	 * @return : the type converter linked to the data base name
	 * is associated to the name, a PException is throwing.
	 */
	public static RdbAdapter getTypeConverter(String dbn) throws RdbAdapterException {
		if (dbn == null || dbn.length() == 0 || dbn.equals(DATABASE_NAME_JDBC))
			return new BasicRdbAdapter();
		String n = null;
		RdbAdapter res = null;
		try {
			n = "org.objectweb.jorm.mapper.rdb.adapter."
					+ Character.toUpperCase(dbn.charAt(0))
					+ dbn.substring(1, dbn.length()).toLowerCase() + "Adapter";
			res = (RdbAdapter) Class.forName(n).newInstance();
		} catch (ClassNotFoundException e) {
			throw new RdbAdapterException(e, "Cannot find an adapter class for this database: " + dbn);
		} catch (ClassCastException e) {
			throw new RdbAdapterException(e, "The specified adpater name is not a RdbAdapter: " + n);
		} catch (Exception e) {
			throw new RdbAdapterException(e, "Impossible to instanciate the following rdb adapter class: " + n);
		}
		return res;
	}

	public static RdbAdapter getTypeConverter(DataSource ds) throws RdbAdapterException {
		Connection conn  = null;
		String adapterName = null;
		try {
			conn = ds.getConnection();
		} catch (SQLException e) {
			throw new RdbAdapterException("Impossible to fetch connection with the specified data source: " + ds);
		}
		try {
			String productName = conn.getMetaData().getDatabaseProductName();
			adapterName = product2adapter.getProperty(productName);
		} catch(SQLException e) {
			adapterName = null;
		} finally {
			try {
				if (conn != null) {
					conn.close();
				}
			} catch (SQLException e) {
			}
		}
		if (adapterName == null) {
			throw new RdbAdapterException("Cannot find automatically a rdb adapter for the data source, please try to define a mapper name like 'rdb.oracle', \ndatasource=" + ds);
		}
		try {
			return (RdbAdapter) Class.forName(adapterName).newInstance();
		} catch (ClassNotFoundException e) {
			throw new RdbAdapterException(e, "The specified adpater name is not a availlable in the classpath: " + adapterName);
		} catch (ClassCastException e) {
			throw new RdbAdapterException(e, "The specified adpater name is not a RdbAdapter: " + adapterName);
		} catch (Exception e) {
			throw new RdbAdapterException(e, "Impossible to instanciate the following rdb adapter class: " + adapterName);
		}
	}

}
