/**
 * Copyright (C) 2004 France Telecom R&D
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */
package org.objectweb.jorm.mapper.rdb.adapter;

import org.objectweb.jorm.mapper.rdb.adapter.api.RdbAdapter;
import org.objectweb.jorm.mapper.rdb.adapter.api.JoinedTable;

import junit.framework.TestCase;
import junit.framework.TestSuite;
import junit.textui.TestRunner;

import java.io.IOException;
import java.io.InputStream;
import java.io.PrintStream;
import java.io.Serializable;
import java.math.BigDecimal;
import java.math.BigInteger;
import java.sql.Connection;
import java.sql.Date;
import java.sql.DriverManager;
import java.sql.PreparedStatement;
import java.sql.ResultSet;
import java.sql.SQLException;
import java.sql.Statement;
import java.util.Properties;
import java.util.StringTokenizer;
import java.util.ArrayList;
import java.util.List;

/**
 * It tests a RdbAdaper. To run this test you have to specify the following
 * system propreties:
 * -Djdbc.driver="driver class name"
 * -Djdbc.url="database url"
 * -Djdbc.user="user name"
 * -Djdbc.passwd="user password"
 * -Dadapter.classname="adapter class name" or -Djdbc.product="database name"
 * Or
 * -Dconfig="list of properties files containing the defintion properties"
 *
 * TODO: test the types: byte[], char[], Serializable
 * TODO: test handleOrderBy(String)
 * TODO: test getSubqueryAlias()
 * TODO: test getIndexedLocateExpression(...)
 * TODO: test getLengthOperator()
 * TODO: test getQuery(String selectClause, List jts, String whereclause);
 * TODO: test getFirstLocateExpression(...)
 * TODO: test getConcatExpression(...)
 * TODO: test getSqlTypeCode(...)
 * TODO: test batch if supportBatchPreparedStatement(...) == true
 *
 * @author S.Chassande-Barrioz
 * @author Tom lane (tgl@redhat.com)
 */
public class TestRdbAdapter extends TestCase {

	public final static String DRIVER = "jdbc.driver";
	public final static String URL = "jdbc.url";
	public final static String ADAPTER_CN = "adapter.classname";
	public final static String PRODUCT = "jdbc.product";
	public final static String USER = "jdbc.user";
	public final static String PASSWD = "jdbc.passwd";
	public final static String CONFIG_FILE_NAME = "config";

	private String drvCN;
	private String user;
	private String url;
	private String passwd = "";
	private PrintStream out = System.out;
	private Connection conn = null;
	private RdbAdapter adapter;
	private boolean debug = true;

	public static void main(String[] args) throws Exception {
		if (args!= null && args.length == 1) {
			System.out.println("TestRdbAdapter usage:");
			System.out.println("=====================");
			System.out.println("The RdbAdapter test is based on the JUnit framework. Its permits to test a JORM");
			System.out.println("RdbAdapter. You have to provide information to acces a database through a JDBC");
			System.out.println("driver. The first way is to specify the following system properties:");
			System.out.println("  -Djdbc.driver=<driver class name>");
			System.out.println("  -Djdbc.url=<database url>");
			System.out.println("  -Djdbc.user=<user name>");
			System.out.println("  -Djdbc.passwd=<user password>");
			System.out.println("  -Dadapter.classname=<adapter class name> or -Djdbc.product=<database name>");
			System.out.println("");
			System.out.println("The second solution is to specify properties files defining the previous ");
			System.out.println("properties such as this");
			System.out.println("-Dconfig=<comma list of .properties files containing the defintion properties>");
			System.out.println("The .properties files will be search in the classpath.");
			System.out.println("");
			System.out.println("IMPORTANT: Do not forget to put in the classpath:");
			System.out.println("            - the JDBC driver,");
			System.out.println("            - the class of the RdbAdapter to test.");
			System.out.println("");
			System.out.println("NOTE: In order to integrate the adapter to Jorm it is adviced to choose the ");
			System.out.println("      same package than the existing (org.objectweb.jorm.mapper.rdb.adapter).");
			System.exit(0);
		}
		TestRdbAdapter t = new TestRdbAdapter();
		t.checkConnection();
		t.close();
		TestRunner.run(new TestSuite(TestRdbAdapter.class));
	}

	public TestRdbAdapter() {
		this("TestRdbAdapter");
	}

	public TestRdbAdapter(String s) {
		super(s);
		configure();
	}

	private void configure() {
		Properties p = System.getProperties();
		String cfn = p.getProperty(CONFIG_FILE_NAME);
		if (cfn != null && cfn.length() > 0) {
			StringTokenizer st = new StringTokenizer(cfn,", ;",false);
			while(st.hasMoreTokens()) {
				cfn = st.nextToken();
				InputStream is = getClass().getClassLoader()
					.getResourceAsStream(cfn);
				if (is != null) {
					try {
						p.load(is);
					} catch (IOException e) {
						fail("Problem during the loading of the configuration file " + cfn);
					}
				} else {
					fail("The configuration file '" + cfn + "' has not been found in the classpath.");
				}
			}
		}
		debug = Boolean.getBoolean("jorm.debug");
		drvCN = p.getProperty(DRIVER);
		if (drvCN == null || drvCN.length() == 0) {
			fail("No property '" + DRIVER + "' defined!");
		}
		url = p.getProperty(URL);
		if (url == null || url.length() == 0) {
			fail("No property '" + URL + "' defined!");
		}
		String acn = p.getProperty(ADAPTER_CN);
		String product = p.getProperty(PRODUCT);
		if (acn !=null && acn.length()>0) {
			try {
				//Load the driver
				adapter = (RdbAdapter) Class.forName(acn).newInstance();
			} catch (Exception e) {
				fail("Impossible to instanciate the adapter: " + acn);
			}
		} else if (product != null && product.length()>0) {
			try {
				adapter = RdbAdapterFactory.getTypeConverter(product);
			} catch (Exception e) {
				fail("Impossible to find an adapter for the database " + product);
			}
		} else {
			fail("No property '" + ADAPTER_CN + "' neither '" + PRODUCT + "' defined!");
		}

		user = p.getProperty(USER);
		passwd = p.getProperty(PASSWD, "");
	}

	protected void setUp() throws Exception {
		checkConnection();
	}

	protected void tearDown() throws Exception {
		close();
	}

	public void checkConnection() throws Exception {
		if (drvCN == null || drvCN.length() == 0) {
			throw new Exception("No driver specified !");
		}
		try {
			//Load the driver
			Class.forName(drvCN);
		} catch (Exception e) {
			out.println("Impossible to load the driver: " + drvCN);
			throw e;
		}
		if (url == null || url.length() == 0) {
			throw new Exception("No url specified !");
		}
		//Fetch a connection
		try {
			if (user == null || user.length() == 0)
				conn = DriverManager.getConnection(url);
			else
				conn = DriverManager.getConnection(url, user, passwd);
		} catch (SQLException e) {
			out.println("Impossible to fetch a connection:" + e.getMessage());
			throw e;
		}
	}

	public void close() {
		if (conn != null) {
			try {
				conn.close();
			} catch (SQLException e) {
			}
		}
	}

	private boolean dropTable(String tableName) {
		Statement s = null;
		try {
			s = conn.createStatement();
			return s.execute("DROP TABLE " + tableName);
		} catch (SQLException e) {
			return false;
		} finally {
			try {
				s.close();
			} catch (SQLException e) {
			}
		}
	}

	private boolean dropSequence(String seqName) {
		Statement s = null;
		try {
			s = conn.createStatement();
			return s.execute("DROP SEQUENCE " + seqName);
		} catch (SQLException e) {
			return false;
		} finally {
			try {
				s.close();
			} catch (SQLException e) {
			}
		}
	}

	private boolean testCreateTable(String tableName, int typeCode, String colType) {
		Statement s = null;
		dropTable(tableName);
		try {
			s = conn.createStatement();
			s.execute(
				"CREATE TABLE " + tableName + "(col1 " + colType + ")");
			return true;
		} catch (SQLException e) {
			fail("create table: " + e.getMessage(), typeCode);
			return false;
		} finally {
			try {
				s.close();
			} catch (SQLException e) {
			}
		}
	}

	public void testSequence() throws Exception {
		conn.setAutoCommit(true);
		String seqName = "JormSeqTest";
		//clean up
		dropSequence(seqName);

		boolean exist;
		try {
			exist = adapter.existSequence(conn, seqName);
			assertTrue("Sequence existence cheking wrong", !exist);
		} catch (SQLException e) {
			fail("Sequence existence cheking fail " + e.getMessage());
		}
		// Sequence creation
		String sqlCreate = null;
		try {
			sqlCreate = adapter.getCreateSequence(seqName);
		} catch (UnsupportedOperationException e) {
			fail("Sequence creation failed: " + e.getMessage());
		}
		Statement s = conn.createStatement();
		s.execute(sqlCreate);

		String sqlNext = adapter.getNextValInSequence(seqName);
		ResultSet rs;
		rs = s.executeQuery(sqlNext);
		if (!rs.next()) {
			fail("Empty resultset for next val in sequence ");
		}
		long l = adapter.getLong(rs, 1, -1);
		assertEquals("Bad sequence value", 1, l);
		rs = s.executeQuery(sqlNext);
		if (!rs.next()) {
			fail("Empty resultset for next val in sequence ");
		}
		assertEquals("Bad sequence value", l+1, adapter.getLong(rs, 1, -1));

		s.close();

		try {
			exist = adapter.existSequence(conn, seqName);
			assertTrue("Sequence existence cheking wrong after creation", exist);
		} catch (SQLException e) {
			fail("Sequence existence cheking fail " + e.getMessage());
		}

		dropSequence(seqName);

		try {
			exist = adapter.existSequence(conn, seqName);
			assertTrue("Sequence existence cheking wrong after deletion", !exist);
		} catch (SQLException e) {
			fail("Sequence existence cheking fail " + e.getMessage());
		}
	}

	public void testTableExistence() throws Exception {
		conn.setAutoCommit(true);
		String tableName = "tableExistest";
		//clean up
		dropTable(tableName);

		boolean exist;
		try {
			exist = adapter.existTable(conn, tableName);
			assertTrue("Table existence cheking wrong", !exist);
		} catch (SQLException e) {
			fail("Table existence cheking fail " + e.getMessage());
		}
		// Table creation
		String colType = "VARCHAR(20)";
		Statement s = null;
		try {
			s = conn.createStatement();
			StringBuffer sb = new StringBuffer();
			sb.append("CREATE TABLE ");
			sb.append(tableName);
			sb.append("(col1 ");
			sb.append(colType);
			sb.append(")");
			s.execute(sb.toString());
		} catch (SQLException e) {
			fail("create table with alias: " + e.getMessage());
		} finally {
			try {
				s.close();
			} catch (SQLException e) {
			}
		}

		try {
			exist = adapter.existTable(conn, tableName);
			assertTrue("Table existence cheking wrong after creation", exist);
		} catch (SQLException e) {
			fail("Table existence cheking fail " + e.getMessage());
		}

		dropTable(tableName);

		try {
			exist = adapter.existTable(conn, tableName);
			assertTrue("Table existence cheking wrong after deletion", !exist);
		} catch (SQLException e) {
			fail("Table existence cheking fail " + e.getMessage());
		}
	}

	public void testAutomaticAdapterDetection() {
		String productName = null;
		try {
			productName = conn.getMetaData().getDatabaseProductName();
		} catch (SQLException e) {
			fail("Impossible to fetch the product name: " + e.getMessage());
		}
		String acn = (String) RdbAdapterFactory.getProductName2Adapter().get(productName);
		assertNotNull("You should register your adapter into the RdbAdapterFactory under the name '"
			+ productName + "'.", acn);
		assertEquals("Current used adapter is not the adapter registered in RdbAdapterFactory the name '"
			+ productName + "'.", acn, adapter.getClass().getName());
	}

	public void testAdapterName() {
		assertNotNull("Null adapter name", adapter.getName());
		assertTrue("Empty adapter name", adapter.getName().length() > 0);
	}

	public void testTableAlias() {
		String tableName = "testTableAlias";
		Statement s = null;
		String colType = null;
		dropTable(tableName);
		try {
			colType = adapter.getSqlType(RdbAdapter.TYPECODE_STRING, false, 20, RdbAdapter.NOSIZE);
			s = conn.createStatement();
			s.execute("CREATE TABLE " + tableName + "(col1 " + colType + ")");
		} catch (Exception e) {
			if (s != null) {
				try {
					s.close();
				} catch (SQLException se) {
				}
			}
			fail("Impossible to create the table: fix problem about String type before");
		}
		try {
			StringBuffer sb = new StringBuffer();
			sb.append("SELECT * from ");
			sb.append(tableName);
			adapter.writeTableAlias("myAlias", sb);
			sb.append(" WHERE col1 = ");
			sb.append(adapter.getValueAsSQLString("String", RdbAdapter.TYPECODE_STRING));
			s.execute(sb.toString());
		} catch (SQLException e) {
			fail("select table with alias: " + e.getMessage());
		} finally {
			if (s != null) {
				try {
					s.close();
				} catch (SQLException e) {
				}
			}
		}
		dropTable(tableName);
	}

	private void testISD(String colType, String tableName,
						 Object val, int typeCode) {
		PreparedStatement ps = null;
		try {
			ps = conn.prepareStatement("INSERT INTO " + tableName + " VALUES (?)");
			assignValue(ps, 1, val, typeCode, colType);
			ps.execute();
		} catch (Exception e) {
			fail("insert: ", e, colType);
		} finally {
			try {
				if (ps != null) {
					ps.close();
				}
			} catch (SQLException e) {
			}
		}

		ResultSet rs = null;
		try {
			ps = conn.prepareStatement("SELECT * from " + tableName + " WHERE col1 = ?");
			assignValue(ps, 1, val, typeCode, colType);
			rs = ps.executeQuery();
			if (!rs.next()) {
				fail("PreparedStatement, select value " + val + ": no result", typeCode);
			}
			Object rval = retrieveValue(rs, 1, typeCode, colType);
			if (!val.equals(rval)) {
				fail("PreparedStatement, select value " + val + ": found other value: " + rval, typeCode);
			}
			if (rs.next()) {
				fail("PreparedStatement, select value " + val + ": more than one result", typeCode);
			}
		} catch (Exception e) {
			fail("PreparedStatement, select: ", e, colType);
		} finally {
			try {
				if (rs != null) {
					rs.close();
				}
			} catch (SQLException e) {
			}
			try {
				if (ps != null) {
					ps.close();
				}
			} catch (SQLException e) {
			}
		}
		Statement s = null;
		try {
			s = conn.createStatement();
			rs = s.executeQuery("SELECT * from " + tableName + " WHERE col1 = "
				+ getValueAsSQLString(val, typeCode));
			if (!rs.next()) {
				fail("Statement, select value " + val + ": no result", typeCode);
			}
			Object rval = retrieveValue(rs, 1, typeCode, colType);
			if (!val.equals(rval)) {
				fail("Statement, select value " + val + ": found other value: " + rval, typeCode);
			}
			if (rs.next()) {
				fail("Statement, select value " + val + ": more than one result", typeCode);
			}
		} catch (Exception e) {
			fail("Statement, select: ", e, colType);
		} finally {
			try {
				if (rs != null) {
					rs.close();
				}
			} catch (SQLException e) {
			}
			try {
				if (s != null) {
					s.close();
				}
			} catch (SQLException e) {
			}
		}

		try {
			s = conn.createStatement();
			int nb = s.executeUpdate("DELETE FROM " + tableName + " WHERE col1 = "
				+ getValueAsSQLString(val, typeCode));
			if (nb != 1) {
				fail("delete value " + val + ": more than one result", typeCode);
			}
		} catch (Exception e) {
			fail("delete: ", e, colType);
		} finally {
			try {
				if (s != null)
					s.close();
			} catch (SQLException e) {
			}
		}
	}

	private String getValueAsSQLString(Object val, int typeCode) {
		switch (typeCode) {
		case RdbAdapter.TYPECODE_BOOLEAN:
			return adapter.getValueAsSQLString(((Boolean) val).booleanValue());
		case RdbAdapter.TYPECODE_BYTE:
			return adapter.getValueAsSQLString(((Byte) val).byteValue());
		case RdbAdapter.TYPECODE_SHORT:
			return adapter.getValueAsSQLString(((Short) val).shortValue());
		case RdbAdapter.TYPECODE_INT:
			return adapter.getValueAsSQLString(((Integer) val).intValue());
		case RdbAdapter.TYPECODE_LONG:
			return adapter.getValueAsSQLString(((Long) val).longValue());
		case RdbAdapter.TYPECODE_FLOAT:
			return adapter.getValueAsSQLString(((Float) val).floatValue());
		case RdbAdapter.TYPECODE_DOUBLE:
			return adapter.getValueAsSQLString(((Double) val).doubleValue());
		case RdbAdapter.TYPECODE_OBJBOOLEAN:
		case RdbAdapter.TYPECODE_OBJBYTE:
		case RdbAdapter.TYPECODE_OBJSHORT:
		case RdbAdapter.TYPECODE_OBJINT:
		case RdbAdapter.TYPECODE_OBJLONG:
		case RdbAdapter.TYPECODE_OBJFLOAT:
		case RdbAdapter.TYPECODE_OBJDOUBLE:
		case RdbAdapter.TYPECODE_OBJCHAR:
		case RdbAdapter.TYPECODE_STRING:
		case RdbAdapter.TYPECODE_DATE:
		case RdbAdapter.TYPECODE_BIGINTEGER:
		case RdbAdapter.TYPECODE_BIGDECIMAL:
		case RdbAdapter.TYPECODE_CHARARRAY:
		case RdbAdapter.TYPECODE_BYTEARRAY:
		case RdbAdapter.TYPECODE_SERIALIZED:
			return adapter.getValueAsSQLString(val, typeCode);
		default:
			return null;
		}
	}

	private void assignValue(PreparedStatement ps,
							 int index,
							 Object value,
							 int typeCode,
							 String colType) throws Exception {
		if (value == null) {
			adapter.setNull(ps, index, adapter.getSqlTypeCode(typeCode));
			return;
		}
		switch (typeCode) {
		case RdbAdapter.TYPECODE_BOOLEAN:
			adapter.setBoolean(ps, index, ((Boolean) value).booleanValue());
			break;
		case RdbAdapter.TYPECODE_CHAR:
			adapter.setChar(ps, index, ((Character) value).charValue());
			break;
		case RdbAdapter.TYPECODE_BYTE:
			adapter.setByte(ps, index, ((Byte) value).byteValue());
			break;
		case RdbAdapter.TYPECODE_SHORT:
			adapter.setShort(ps, index, ((Short) value).shortValue());
			break;
		case RdbAdapter.TYPECODE_INT:
			adapter.setInt(ps, index, ((Integer) value).intValue());
			break;
		case RdbAdapter.TYPECODE_LONG:
			adapter.setLong(ps, index, ((Long) value).longValue());
			break;
		case RdbAdapter.TYPECODE_FLOAT:
			adapter.setFloat(ps, index, ((Float) value).floatValue());
			break;
		case RdbAdapter.TYPECODE_DOUBLE:
			adapter.setDouble(ps, index, ((Double) value).doubleValue());
			break;
		case RdbAdapter.TYPECODE_OBJBOOLEAN:
			adapter.setOboolean(ps, index, (Boolean) value);
			break;
		case RdbAdapter.TYPECODE_OBJCHAR:
			adapter.setOchar(ps, index, (Character) value);
			break;
		case RdbAdapter.TYPECODE_OBJBYTE:
			adapter.setObyte(ps, index, (Byte) value);
			break;
		case RdbAdapter.TYPECODE_OBJSHORT:
			adapter.setOshort(ps, index, (Short) value);
			break;
		case RdbAdapter.TYPECODE_OBJINT:
			adapter.setOint(ps, index, (Integer) value);
			break;
		case RdbAdapter.TYPECODE_OBJLONG:
			adapter.setOlong(ps, index, (Long) value);
			break;
		case RdbAdapter.TYPECODE_OBJFLOAT:
			adapter.setOfloat(ps, index, (Float) value);
			break;
		case RdbAdapter.TYPECODE_OBJDOUBLE:
			adapter.setOdouble(ps, index, (Double) value);
			break;
		case RdbAdapter.TYPECODE_STRING:
			adapter.setString(ps, index, (String) value);
			break;
		case RdbAdapter.TYPECODE_DATE:
			adapter.setDate(ps, index, (Date) value, colType);
			break;
		case RdbAdapter.TYPECODE_CHARARRAY:
			adapter.setCharArray(ps, index, (char[]) value);
			break;
		case RdbAdapter.TYPECODE_BYTEARRAY:
			adapter.setByteArray(ps, index, (byte[]) value);
			break;
		case RdbAdapter.TYPECODE_SERIALIZED:
			adapter.setSerialized(ps, index, (Serializable) value);
			break;
		case RdbAdapter.TYPECODE_BIGINTEGER:
			adapter.setBigInteger(ps, index, (BigInteger) value);
			break;
		case RdbAdapter.TYPECODE_BIGDECIMAL:
			adapter.setBigDecimal(ps, index, (BigDecimal) value);
			break;
		default:
			throw new Exception("Unmanaged type: " + RdbAdapter.TYPE_NAMES[typeCode]);
		}
	}

	private Object retrieveValue(ResultSet rs,
								 int index,
								 int typeCode,
								 String colType) throws Exception {
		switch (typeCode) {
		case RdbAdapter.TYPECODE_BOOLEAN:
			return new Boolean(adapter.getBoolean(rs, index, false));
		case RdbAdapter.TYPECODE_CHAR:
			return new Character(adapter.getChar(rs, index, (char) 0));
		case RdbAdapter.TYPECODE_BYTE:
			return new Byte(adapter.getByte(rs, index, (byte) 0));
		case RdbAdapter.TYPECODE_SHORT:
			return new Short(adapter.getShort(rs, index, (short) 0));
		case RdbAdapter.TYPECODE_INT:
			return new Integer(adapter.getInt(rs, index, 0));
		case RdbAdapter.TYPECODE_LONG:
			return new Long(adapter.getLong(rs, index, 0));
		case RdbAdapter.TYPECODE_FLOAT:
			return new Float(adapter.getFloat(rs, index, 0));
		case RdbAdapter.TYPECODE_DOUBLE:
			return new Double(adapter.getDouble(rs, index, 0));
		case RdbAdapter.TYPECODE_OBJBOOLEAN:
			return adapter.getOboolean(rs, index, null);
		case RdbAdapter.TYPECODE_OBJCHAR:
			return adapter.getOchar(rs, index, null);
		case RdbAdapter.TYPECODE_OBJBYTE:
			return adapter.getObyte(rs, index, null);
		case RdbAdapter.TYPECODE_OBJSHORT:
			return adapter.getOshort(rs, index, null);
		case RdbAdapter.TYPECODE_OBJINT:
			return adapter.getOint(rs, index, null);
		case RdbAdapter.TYPECODE_OBJLONG:
			return adapter.getOlong(rs, index, null);
		case RdbAdapter.TYPECODE_OBJFLOAT:
			return adapter.getOfloat(rs, index, null);
		case RdbAdapter.TYPECODE_OBJDOUBLE:
			return adapter.getOdouble(rs, index, null);
		case RdbAdapter.TYPECODE_STRING:
			return adapter.getString(rs, index, null);
		case RdbAdapter.TYPECODE_DATE:
			return adapter.getDate(rs, index, colType, null);
		case RdbAdapter.TYPECODE_CHARARRAY:
			return adapter.getCharArray(rs, index, null);
		case RdbAdapter.TYPECODE_BYTEARRAY:
			return adapter.getByteArray(rs, index, null);
		case RdbAdapter.TYPECODE_SERIALIZED:
			return adapter.getSerialized(rs, index, null);
		case RdbAdapter.TYPECODE_BIGINTEGER:
			return adapter.getBigInteger(rs, index, null);
		case RdbAdapter.TYPECODE_BIGDECIMAL:
			return adapter.getBigDecimal(rs, index, null);
		default:
			throw new Exception("Unmanaged type: " + RdbAdapter.TYPE_NAMES[typeCode]);
		}
	}


	public void testBoolean() {
		testType(RdbAdapter.TYPECODE_BOOLEAN,
			new Object[]{
				Boolean.TRUE,
				Boolean.FALSE
			}, RdbAdapter.NOSIZE, RdbAdapter.NOSIZE);
	}

	public void testObjBoolean() {
		testType(RdbAdapter.TYPECODE_OBJBOOLEAN,
			new Object[]{
				Boolean.TRUE,
				Boolean.FALSE
			}, RdbAdapter.NOSIZE, RdbAdapter.NOSIZE);
	}

	public void testByte() {
		testType(RdbAdapter.TYPECODE_BYTE,
			new Object[]{
				new Byte((byte) 0),
				new Byte((byte) 1),
				new Byte((byte) 254)
			}, RdbAdapter.NOSIZE, RdbAdapter.NOSIZE);
	}

	public void testObjByte() {
		testType(RdbAdapter.TYPECODE_OBJBYTE,
			new Object[]{
				new Byte((byte) 0),
				new Byte((byte) 1),
				new Byte((byte) 254)
			}, RdbAdapter.NOSIZE, RdbAdapter.NOSIZE);
	}

	public void testShort() {
		testType(RdbAdapter.TYPECODE_SHORT,
			new Object[]{
				new Short((short) 0),
				new Short((short) 1),
				new Short((short) 254)
			}, RdbAdapter.NOSIZE, RdbAdapter.NOSIZE);
	}

	public void testObjShort() {
		testType(RdbAdapter.TYPECODE_OBJSHORT,
			new Object[]{
				new Short((short) 0),
				new Short((short) 1),
				new Short((short) 254)
			}, RdbAdapter.NOSIZE, RdbAdapter.NOSIZE);
	}

	public void testInt() {
		testType(RdbAdapter.TYPECODE_INT,
			new Object[]{
				new Integer(0),
				new Integer(1),
				new Integer(254)
			}, RdbAdapter.NOSIZE, RdbAdapter.NOSIZE);
	}

	public void testObjInt() {
		testType(RdbAdapter.TYPECODE_OBJINT,
			new Object[]{
				new Integer(0),
				new Integer(1),
				new Integer(254)
			}, RdbAdapter.NOSIZE, RdbAdapter.NOSIZE);
	}

	public void testLong() {
		testType(RdbAdapter.TYPECODE_LONG,
			new Object[]{
				new Long(0),
				new Long(1),
				new Long(254)
			}, RdbAdapter.NOSIZE, RdbAdapter.NOSIZE);
	}

	public void testObjLong() {
		testType(RdbAdapter.TYPECODE_OBJLONG,
			new Object[]{
				new Long(0),
				new Long(1),
				new Long(254)
			}, RdbAdapter.NOSIZE, RdbAdapter.NOSIZE);
	}

	public void testFloat() {
		testType(RdbAdapter.TYPECODE_FLOAT,
			new Object[]{
				new Float(0.0),
				new Float(1),
				new Float(254)
			}, RdbAdapter.NOSIZE, RdbAdapter.NOSIZE);
	}

	public void testObjFloat() {
		testType(RdbAdapter.TYPECODE_OBJFLOAT,
			new Object[]{
				new Float(0.0),
				new Float(1),
				new Float(254)
			}, RdbAdapter.NOSIZE, RdbAdapter.NOSIZE);
	}

	public void testDouble() {
		testType(RdbAdapter.TYPECODE_DOUBLE,
			new Object[]{
				new Double(0.0),
				new Double(1),
				new Double(254)
			}, RdbAdapter.NOSIZE, RdbAdapter.NOSIZE);
	}

	public void testObjDouble() {
		testType(RdbAdapter.TYPECODE_OBJDOUBLE,
			new Object[]{
				new Double(0.0),
				new Double(1),
				new Double(254)
			}, RdbAdapter.NOSIZE, RdbAdapter.NOSIZE);
	}

	public void testDate() {
		testType(RdbAdapter.TYPECODE_DATE,
			new Object[]{
			}, RdbAdapter.NOSIZE, RdbAdapter.NOSIZE);
	}

	public void testString() {
		testType(RdbAdapter.TYPECODE_STRING,
			new Object[]{
				"",
				"abcd"
			}, RdbAdapter.NOSIZE, RdbAdapter.NOSIZE);
	}

	public void testStringWithSize() {
		testType(RdbAdapter.TYPECODE_STRING,
			new Object[]{
				"",
				"abcd"
			}, 24, RdbAdapter.NOSIZE);
	}

	public void testBigDecimal() {
		testType(RdbAdapter.TYPECODE_BIGDECIMAL,
			new Object[]{
				new BigDecimal("00000.0000000000"),
				new BigDecimal("145.0123456789")
			}, 31, 10);
	}

	public void testBigInteger() {
		testType(RdbAdapter.TYPECODE_BIGINTEGER,
			new Object[]{
				new BigInteger("0000000000000000000"),
				new BigInteger("123456789012345")
			}, 20, RdbAdapter.NOSIZE);
	}

	private String getTestQuery(boolean rangeStart, boolean rangeSize) {
		JoinedTable jt = new JoinedTable("t1");
		List jts = new ArrayList();
		jts.add(jt);
		return adapter.getQuery("a, b", jts, "a < 12", rangeStart, rangeSize);
	}

	public void testQueryNoRange() {
		assertEquals("Incorrect query", getTestQuery(false, false), "SELECT a, b FROM t1 WHERE a < 12");
	}

	public void testQueryOffsetAndLimit() {
		String result = getTestQuery(true, true);
		if (adapter.getName().equals("postgres")) {
			assertEquals("Incorrect query", result, "SELECT a, b FROM t1 WHERE a < 12 OFFSET ? LIMIT ?");
		} else if (adapter.getName().equals("mysql")) {
			assertEquals("Incorrect query", result, "SELECT a, b FROM t1 WHERE a < 12 LIMIT ?, ?");
		} else if (adapter.getName().equals("oracle")) {
			assertEquals("Incorrect query", result, "SELECT * FROM (SELECT a, b, ROWNUM as RNUM_ FROM t1 WHERE a < 12) WHERE RNUM_ > ? AND RNUM_ <= ? + ?");
		} else if (adapter.getName().equals("hsql")) {
			assertEquals("Incorrect query", result, "SELECT  LIMIT ? ? a, b FROM t1 WHERE a < 12");
		} else {
			assertEquals("Incorrect query", result, "SELECT a, b FROM t1 WHERE a < 12");
		}
	}

	public void testQueryLimit() {
		String result = getTestQuery(false, true);
		if (adapter.getName().equals("postgres")) {
			assertEquals("Incorrect query", result, "SELECT a, b FROM t1 WHERE a < 12 LIMIT ?");
		} else if (adapter.getName().equals("mysql")) {
			assertEquals("Incorrect query", result, "SELECT a, b FROM t1 WHERE a < 12 LIMIT 0, ?");
		} else if (adapter.getName().equals("oracle")) {
			assertEquals("Incorrect query", result, "SELECT * FROM (SELECT a, b, ROWNUM as RNUM_ FROM t1 WHERE a < 12) WHERE RNUM_ <= ?");
		} else if (adapter.getName().equals("hsql")) {
			assertEquals("Incorrect query", result, "SELECT  LIMIT 0 ? a, b FROM t1 WHERE a < 12");
		} else {
			assertEquals("Incorrect query", result, "SELECT a, b FROM t1 WHERE a < 12");
		}
	}

	public void testQueryOffset() {
		String result = getTestQuery(true, false);
		if (adapter.getName().equals("postgres")) {
			assertEquals("Incorrect query", result, "SELECT a, b FROM t1 WHERE a < 12 OFFSET ?");
		} else if (adapter.getName().equals("mysql")) {
			assertEquals("Incorrect query", result, "SELECT a, b FROM t1 WHERE a < 12 LIMIT ?, 18446744073709551615");
		} else if (adapter.getName().equals("oracle")) {
			assertEquals("Incorrect query", result, "SELECT * FROM (SELECT a, b, ROWNUM as RNUM_ FROM t1 WHERE a < 12) WHERE RNUM_ > ?");
		} else if (adapter.getName().equals("hsql")) {
			assertEquals("Incorrect query", result, "SELECT  LIMIT 3 0 a, b FROM t1 WHERE a < 12");
		} else {
			assertEquals("Incorrect query", result, "SELECT a, b FROM t1 WHERE a < 12");
		}
	}

	private void testType(int typeCode, Object[] values, int size, int scale) {
		String tableName = "Jorm" + RdbAdapter.TYPE_NAMES[typeCode].replace('.','_');
		String colType;
		try {
			colType = adapter.getSqlType(typeCode, false, size, scale);
		} catch (Exception e) {
			fail("Impossible to fetch the SQL column type",
				e, RdbAdapter.TYPE_NAMES[typeCode]);
			return;
		}
		testCreateTable(tableName, typeCode, colType);
		if (testCreateTable(tableName, typeCode, colType)) {
			for (int i = 0; i < values.length; i++) {
				testISD(colType, tableName, values[i], typeCode);
			}
		}
		dropTable(tableName);
	}


	public void fail(String msg, Throwable t, String colType) {
		if (debug) {
			fail(colType + ": " + msg);
			t.printStackTrace(out);
		} else
			fail(colType + ": " + msg + t.getMessage());
	}

	public void fail(String msg, int typeCode) {
		fail(RdbAdapter.TYPE_NAMES[typeCode] + ": " + msg);
	}
}
